@@ -12,18 +12,12 @@
 #include <string.h>
 #include <time.h>
 #include <assert.h>
+#include <unistd.h>
 
 /* Forward Declarations */
 static void item_link_q(item *it);
 static void item_unlink_q(item *it);
 
-/*
- * We only reposition items in the LRU queue if they haven't been repositioned
- * in this many seconds. That saves us from churning on frequently-accessed
- * items.
- */
-#define ITEM_UPDATE_INTERVAL 60
-
 #define LARGEST_ID POWER_LARGEST
 typedef struct {
     uint64_t evicted;
@@ -34,13 +28,22 @@ typedef struct {
     uint64_t tailrepairs;
     uint64_t expired_unfetched;
     uint64_t evicted_unfetched;
+    uint64_t crawler_reclaimed;
+    uint64_t lrutail_reflocked;
 } itemstats_t;
 
 static item *heads[LARGEST_ID];
 static item *tails[LARGEST_ID];
+static crawler crawlers[LARGEST_ID];
 static itemstats_t itemstats[LARGEST_ID];
 static unsigned int sizes[LARGEST_ID];
 
+static int crawler_count = 0;
+static volatile int do_run_lru_crawler_thread = 0;
+static int lru_crawler_initialized = 0;
+static pthread_mutex_t lru_crawler_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t  lru_crawler_cond = PTHREAD_COND_INITIALIZER;
+
 void item_stats_reset(void) {
     mutex_lock(&cache_lock);
     memset(itemstats, 0, sizeof(itemstats));
@@ -103,35 +106,55 @@ item *do_item_alloc(char *key, const size_t nkey, const int flags,
     mutex_lock(&cache_lock);
     /* do a quick check if we have any expired items in the tail.. */
     int tries = 5;
+    /* Avoid hangs if a slab has nothing but refcounted stuff in it. */
+    int tries_lrutail_reflocked = 1000;
     int tried_alloc = 0;
     item *search;
+    item *next_it;
     void *hold_lock = NULL;
     rel_time_t oldest_live = settings.oldest_live;
 
     search = tails[id];
     /* We walk up *only* for locked items. Never searching for expired.
      * Waste of CPU for almost all deployments */
-    for (; tries > 0 && search != NULL; tries--, search=search->prev) {
-        uint32_t hv = hash(ITEM_key(search), search->nkey, 0);
+    for (; tries > 0 && search != NULL; tries--, search=next_it) {
+        /* we might relink search mid-loop, so search->prev isn't reliable */
+        next_it = search->prev;
+        if (search->nbytes == 0 && search->nkey == 0 && search->it_flags == 1) {
+            /* We are a crawler, ignore it. */
+            tries++;
+            continue;
+        }
+        uint32_t hv = hash(ITEM_key(search), search->nkey);
         /* Attempt to hash item lock the "search" item. If locked, no
          * other callers can incr the refcount
          */
-        /* FIXME: I think we need to mask the hv here for comparison? */
-        if (hv != cur_hv && (hold_lock = item_trylock(hv)) == NULL)
+        /* Don't accidentally grab ourselves, or bail if we can't quicklock */
+        if (hv == cur_hv || (hold_lock = item_trylock(hv)) == NULL)
             continue;
         /* Now see if the item is refcount locked */
         if (refcount_incr(&search->refcount) != 2) {
+            /* Avoid pathological case with ref'ed items in tail */
+            do_item_update_nolock(search);
+            tries_lrutail_reflocked--;
+            tries++;
             refcount_decr(&search->refcount);
+            itemstats[id].lrutail_reflocked++;
             /* Old rare bug could cause a refcount leak. We haven't seen
              * it in years, but we leave this code in to prevent failures
              * just in case */
-            if (search->time + settings.tail_repair_time < current_time) {
+            if (settings.tail_repair_time &&
+                    search->time + settings.tail_repair_time < current_time) {
                 itemstats[id].tailrepairs++;
                 search->refcount = 1;
                 do_item_unlink_nolock(search, hv);
             }
             if (hold_lock)
                 item_trylock_unlock(hold_lock);
+
+            if (tries_lrutail_reflocked < 1)
+                break;
+
             continue;
         }
 
@@ -352,6 +375,21 @@ void do_item_remove(item *it) {
     }
 }
 
+/* Copy/paste to avoid adding two extra branches for all common calls, since
+ * _nolock is only used in an uncommon case. */
+void do_item_update_nolock(item *it) {
+    MEMCACHED_ITEM_UPDATE(ITEM_key(it), it->nkey, it->nbytes);
+    if (it->time < current_time - ITEM_UPDATE_INTERVAL) {
+        assert((it->it_flags & ITEM_SLABBED) == 0);
+
+        if ((it->it_flags & ITEM_LINKED) != 0) {
+            item_unlink_q(it);
+            it->time = current_time;
+            item_link_q(it);
+        }
+    }
+}
+
 void do_item_update(item *it) {
     MEMCACHED_ITEM_UPDATE(ITEM_key(it), it->nkey, it->nbytes);
     if (it->time < current_time - ITEM_UPDATE_INTERVAL) {
@@ -395,6 +433,10 @@ char *do_item_cachedump(const unsigned int slabs_clsid, const unsigned int limit
 
     while (it != NULL && (limit == 0 || shown < limit)) {
         assert(it->nkey <= KEY_MAX_LENGTH);
+        if (it->nbytes == 0 && it->nkey == 0) {
+            it = it->next;
+            continue;
+        }
         /* Copy the key since it may not be null-terminated in the struct */
         strncpy(key_temp, ITEM_key(it), it->nkey);
         key_temp[it->nkey] = 0x00; /* terminate */
@@ -434,6 +476,8 @@ void do_item_stats_totals(ADD_STAT add_stats, void *c) {
         totals.evicted_unfetched += itemstats[i].evicted_unfetched;
         totals.evicted += itemstats[i].evicted;
         totals.reclaimed += itemstats[i].reclaimed;
+        totals.crawler_reclaimed += itemstats[i].crawler_reclaimed;
+        totals.lrutail_reflocked += itemstats[i].lrutail_reflocked;
     }
     APPEND_STAT("expired_unfetched", "%llu",
                 (unsigned long long)totals.expired_unfetched);
@@ -443,6 +487,10 @@ void do_item_stats_totals(ADD_STAT add_stats, void *c) {
                 (unsigned long long)totals.evicted);
     APPEND_STAT("reclaimed", "%llu",
                 (unsigned long long)totals.reclaimed);
+    APPEND_STAT("crawler_reclaimed", "%llu",
+                (unsigned long long)totals.crawler_reclaimed);
+    APPEND_STAT("lrutail_reflocked", "%llu",
+                (unsigned long long)totals.lrutail_reflocked);
 }
 
 void do_item_stats(ADD_STAT add_stats, void *c) {
@@ -475,6 +523,10 @@ void do_item_stats(ADD_STAT add_stats, void *c) {
                                 "%llu", (unsigned long long)itemstats[i].expired_unfetched);
             APPEND_NUM_FMT_STAT(fmt, i, "evicted_unfetched",
                                 "%llu", (unsigned long long)itemstats[i].evicted_unfetched);
+            APPEND_NUM_FMT_STAT(fmt, i, "crawler_reclaimed",
+                                "%llu", (unsigned long long)itemstats[i].crawler_reclaimed);
+            APPEND_NUM_FMT_STAT(fmt, i, "lrutail_reflocked",
+                                "%llu", (unsigned long long)itemstats[i].lrutail_reflocked);
         }
     }
 
@@ -600,10 +652,11 @@ void do_item_flush_expired(void) {
          * The oldest_live checking will auto-expire the remaining items.
          */
         for (iter = heads[i]; iter != NULL; iter = next) {
-            if (iter->time >= settings.oldest_live) {
+            /* iter->time of 0 are magic objects. */
+            if (iter->time != 0 && iter->time >= settings.oldest_live) {
                 next = iter->next;
                 if ((iter->it_flags & ITEM_SLABBED) == 0) {
-                    do_item_unlink_nolock(iter, hash(ITEM_key(iter), iter->nkey, 0));
+                    do_item_unlink_nolock(iter, hash(ITEM_key(iter), iter->nkey));
                 }
             } else {
                 /* We've hit the first old item. Continue to the next queue. */
@@ -612,3 +665,307 @@ void do_item_flush_expired(void) {
         }
     }
 }
+
+static void crawler_link_q(item *it) { /* item is the new tail */
+    item **head, **tail;
+    assert(it->slabs_clsid < LARGEST_ID);
+    assert(it->it_flags == 1);
+    assert(it->nbytes == 0);
+
+    head = &heads[it->slabs_clsid];
+    tail = &tails[it->slabs_clsid];
+    assert(*tail != 0);
+    assert(it != *tail);
+    assert((*head && *tail) || (*head == 0 && *tail == 0));
+    it->prev = *tail;
+    it->next = 0;
+    if (it->prev) {
+        assert(it->prev->next == 0);
+        it->prev->next = it;
+    }
+    *tail = it;
+    if (*head == 0) *head = it;
+    return;
+}
+
+static void crawler_unlink_q(item *it) {
+    item **head, **tail;
+    assert(it->slabs_clsid < LARGEST_ID);
+    head = &heads[it->slabs_clsid];
+    tail = &tails[it->slabs_clsid];
+
+    if (*head == it) {
+        assert(it->prev == 0);
+        *head = it->next;
+    }
+    if (*tail == it) {
+        assert(it->next == 0);
+        *tail = it->prev;
+    }
+    assert(it->next != it);
+    assert(it->prev != it);
+
+    if (it->next) it->next->prev = it->prev;
+    if (it->prev) it->prev->next = it->next;
+    return;
+}
+
+/* This is too convoluted, but it's a difficult shuffle. Try to rewrite it
+ * more clearly. */
+static item *crawler_crawl_q(item *it) {
+    item **head, **tail;
+    assert(it->it_flags == 1);
+    assert(it->nbytes == 0);
+    assert(it->slabs_clsid < LARGEST_ID);
+    head = &heads[it->slabs_clsid];
+    tail = &tails[it->slabs_clsid];
+
+    /* We've hit the head, pop off */
+    if (it->prev == 0) {
+        assert(*head == it);
+        if (it->next) {
+            *head = it->next;
+            assert(it->next->prev == it);
+            it->next->prev = 0;
+        }
+        return NULL; /* Done */
+    }
+
+    /* Swing ourselves in front of the next item */
+    /* NB: If there is a prev, we can't be the head */
+    assert(it->prev != it);
+    if (it->prev) {
+        if (*head == it->prev) {
+            /* Prev was the head, now we're the head */
+            *head = it;
+        }
+        if (*tail == it) {
+            /* We are the tail, now they are the tail */
+            *tail = it->prev;
+        }
+        assert(it->next != it);
+        if (it->next) {
+            assert(it->prev->next == it);
+            it->prev->next = it->next;
+            it->next->prev = it->prev;
+        } else {
+            /* Tail. Move this above? */
+            it->prev->next = 0;
+        }
+        /* prev->prev's next is it->prev */
+        it->next = it->prev;
+        it->prev = it->next->prev;
+        it->next->prev = it;
+        /* New it->prev now, if we're not at the head. */
+        if (it->prev) {
+            it->prev->next = it;
+        }
+    }
+    assert(it->next != it);
+    assert(it->prev != it);
+
+    return it->next; /* success */
+}
+
+/* I pulled this out to make the main thread clearer, but it reaches into the
+ * main thread's values too much. Should rethink again.
+ */
+static void item_crawler_evaluate(item *search, uint32_t hv, int i) {
+    rel_time_t oldest_live = settings.oldest_live;
+    if ((search->exptime != 0 && search->exptime < current_time)
+        || (search->time <= oldest_live && oldest_live <= current_time)) {
+        itemstats[i].crawler_reclaimed++;
+
+        if (settings.verbose > 1) {
+            int ii;
+            char *key = ITEM_key(search);
+            fprintf(stderr, "LRU crawler found an expired item (flags: %d, slab: %d): ",
+                search->it_flags, search->slabs_clsid);
+            for (ii = 0; ii < search->nkey; ++ii) {
+                fprintf(stderr, "%c", key[ii]);
+            }
+            fprintf(stderr, "\n");
+        }
+        if ((search->it_flags & ITEM_FETCHED) == 0) {
+            itemstats[i].expired_unfetched++;
+        }
+        do_item_unlink_nolock(search, hv);
+        do_item_remove(search);
+        assert(search->slabs_clsid == 0);
+    } else {
+        refcount_decr(&search->refcount);
+    }
+}
+
+static void *item_crawler_thread(void *arg) {
+    int i;
+
+    pthread_mutex_lock(&lru_crawler_lock);
+    if (settings.verbose > 2)
+        fprintf(stderr, "Starting LRU crawler background thread\n");
+    while (do_run_lru_crawler_thread) {
+    pthread_cond_wait(&lru_crawler_cond, &lru_crawler_lock);
+
+    while (crawler_count) {
+        item *search = NULL;
+        void *hold_lock = NULL;
+
+        for (i = 0; i < LARGEST_ID; i++) {
+            if (crawlers[i].it_flags != 1) {
+                continue;
+            }
+            pthread_mutex_lock(&cache_lock);
+            search = crawler_crawl_q((item *)&crawlers[i]);
+            if (search == NULL ||
+                (crawlers[i].remaining && --crawlers[i].remaining < 1)) {
+                if (settings.verbose > 2)
+                    fprintf(stderr, "Nothing left to crawl for %d\n", i);
+                crawlers[i].it_flags = 0;
+                crawler_count--;
+                crawler_unlink_q((item *)&crawlers[i]);
+                pthread_mutex_unlock(&cache_lock);
+                continue;
+            }
+            uint32_t hv = hash(ITEM_key(search), search->nkey);
+            /* Attempt to hash item lock the "search" item. If locked, no
+             * other callers can incr the refcount
+             */
+            if ((hold_lock = item_trylock(hv)) == NULL) {
+                pthread_mutex_unlock(&cache_lock);
+                continue;
+            }
+            /* Now see if the item is refcount locked */
+            if (refcount_incr(&search->refcount) != 2) {
+                refcount_decr(&search->refcount);
+                if (hold_lock)
+                    item_trylock_unlock(hold_lock);
+                pthread_mutex_unlock(&cache_lock);
+                continue;
+            }
+
+            /* Frees the item or decrements the refcount. */
+            /* Interface for this could improve: do the free/decr here
+             * instead? */
+            item_crawler_evaluate(search, hv, i);
+
+            if (hold_lock)
+                item_trylock_unlock(hold_lock);
+            pthread_mutex_unlock(&cache_lock);
+
+            if (settings.lru_crawler_sleep)
+                usleep(settings.lru_crawler_sleep);
+        }
+    }
+    if (settings.verbose > 2)
+        fprintf(stderr, "LRU crawler thread sleeping\n");
+    STATS_LOCK();
+    stats.lru_crawler_running = false;
+    STATS_UNLOCK();
+    }
+    pthread_mutex_unlock(&lru_crawler_lock);
+    if (settings.verbose > 2)
+        fprintf(stderr, "LRU crawler thread stopping\n");
+
+    return NULL;
+}
+
+static pthread_t item_crawler_tid;
+
+int stop_item_crawler_thread(void) {
+    int ret;
+    pthread_mutex_lock(&lru_crawler_lock);
+    do_run_lru_crawler_thread = 0;
+    pthread_cond_signal(&lru_crawler_cond);
+    pthread_mutex_unlock(&lru_crawler_lock);
+    if ((ret = pthread_join(item_crawler_tid, NULL)) != 0) {
+        fprintf(stderr, "Failed to stop LRU crawler thread: %s\n", strerror(ret));
+        return -1;
+    }
+    settings.lru_crawler = false;
+    return 0;
+}
+
+int start_item_crawler_thread(void) {
+    int ret;
+
+    if (settings.lru_crawler)
+        return -1;
+    pthread_mutex_lock(&lru_crawler_lock);
+    do_run_lru_crawler_thread = 1;
+    settings.lru_crawler = true;
+    if ((ret = pthread_create(&item_crawler_tid, NULL,
+        item_crawler_thread, NULL)) != 0) {
+        fprintf(stderr, "Can't create LRU crawler thread: %s\n",
+            strerror(ret));
+        pthread_mutex_unlock(&lru_crawler_lock);
+        return -1;
+    }
+    pthread_mutex_unlock(&lru_crawler_lock);
+
+    return 0;
+}
+
+enum crawler_result_type lru_crawler_crawl(char *slabs) {
+    char *b = NULL;
+    uint32_t sid = 0;
+    uint8_t tocrawl[POWER_LARGEST];
+    if (pthread_mutex_trylock(&lru_crawler_lock) != 0) {
+        return CRAWLER_RUNNING;
+    }
+    pthread_mutex_lock(&cache_lock);
+
+    if (strcmp(slabs, "all") == 0) {
+        for (sid = 0; sid < LARGEST_ID; sid++) {
+            tocrawl[sid] = 1;
+        }
+    } else {
+        for (char *p = strtok_r(slabs, ",", &b);
+             p != NULL;
+             p = strtok_r(NULL, ",", &b)) {
+
+            if (!safe_strtoul(p, &sid) || sid < POWER_SMALLEST
+                    || sid > POWER_LARGEST) {
+                pthread_mutex_unlock(&cache_lock);
+                pthread_mutex_unlock(&lru_crawler_lock);
+                return CRAWLER_BADCLASS;
+            }
+            tocrawl[sid] = 1;
+        }
+    }
+
+    for (sid = 0; sid < LARGEST_ID; sid++) {
+        if (tocrawl[sid] != 0 && tails[sid] != NULL) {
+            if (settings.verbose > 2)
+                fprintf(stderr, "Kicking LRU crawler off for slab %d\n", sid);
+            crawlers[sid].nbytes = 0;
+            crawlers[sid].nkey = 0;
+            crawlers[sid].it_flags = 1; /* For a crawler, this means enabled. */
+            crawlers[sid].next = 0;
+            crawlers[sid].prev = 0;
+            crawlers[sid].time = 0;
+            crawlers[sid].remaining = settings.lru_crawler_tocrawl;
+            crawlers[sid].slabs_clsid = sid;
+            crawler_link_q((item *)&crawlers[sid]);
+            crawler_count++;
+        }
+    }
+    pthread_mutex_unlock(&cache_lock);
+    pthread_cond_signal(&lru_crawler_cond);
+    STATS_LOCK();
+    stats.lru_crawler_running = true;
+    STATS_UNLOCK();
+    pthread_mutex_unlock(&lru_crawler_lock);
+    return CRAWLER_OK;
+}
+
+int init_lru_crawler(void) {
+    if (lru_crawler_initialized == 0) {
+        if (pthread_cond_init(&lru_crawler_cond, NULL) != 0) {
+            fprintf(stderr, "Can't initialize lru crawler condition\n");
+            return -1;
+        }
+        pthread_mutex_init(&lru_crawler_lock, NULL);
+        lru_crawler_initialized = 1;
+    }
+    return 0;
+}
